@@ -1,7 +1,7 @@
 # ----------------------------------------------------------------------------
 # pyglet
 # Copyright (c) 2006-2008 Alex Holkner
-# Copyright (c) 2008-2021 pyglet contributors
+# Copyright (c) 2008-2022 pyglet contributors
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -38,19 +38,19 @@
 Measuring time
 ==============
 
-The `tick` and `get_fps` functions can be used in conjunction to fulfil most
+The `tick` and `get_frequency` functions can be used in conjunction to fulfil most
 games' basic requirements::
 
     from pyglet import clock
     while True:
         dt = clock.tick()
         # ... update and render ...
-        print(f"FPS is {clock.get_fps()}")
+        print(f"FPS is {clock.get_frequency()}")
 
 The ``dt`` value returned gives the number of seconds (as a float) since the
 last "tick".
 
-The `get_fps` function averages the framerate over a sliding window of
+The `get_frequency` function averages the framerate over a sliding window of
 approximately 1 second.  (You can calculate the instantaneous framerate by
 taking the reciprocal of ``dt``).
 
@@ -111,13 +111,13 @@ Multiple and derived clocks potentially allow you to separate "game-time" and
 of the system clock.
 """
 
-import time
+import time as _time
 
-from operator import attrgetter
-from heapq import heappush, heappop, heappushpop
-from collections import deque
-
-from pyglet import compat_platform
+from heapq import heappop as _heappop
+from heapq import heappush as _heappush
+from heapq import heappushpop as _heappushpop
+from operator import attrgetter as _attrgetter
+from collections import deque as _deque
 
 
 class _ScheduledItem:
@@ -152,22 +152,6 @@ class Clock:
 
     It is also used for calling scheduled functions.
     """
-
-    #: The minimum amount of time in seconds this clock will attempt to sleep
-    #: for when framerate limiting.  Higher values will increase the
-    #: accuracy of the limiting but also increase CPU usage while
-    #: busy-waiting.  Lower values mean the process sleeps more often, but is
-    #: prone to over-sleep and run at a potentially lower or uneven framerate
-    #: than desired.
-
-    #: On Windows, MIN_SLEEP is larger because the default timer resolution
-    #: is set by default to 15 .6 ms.
-    MIN_SLEEP = 0.008 if compat_platform in ('win32', 'cygwin') else 0.005
-
-    #: The amount of time in seconds this clock subtracts from sleep values
-    #: to compensate for lazy operating systems.
-    SLEEP_UNDERSHOOT = MIN_SLEEP - 0.001
-
     # List of functions to call every tick.
     _schedule_items = None
 
@@ -177,7 +161,7 @@ class Clock:
     # If True, a sleep(0) is inserted on every tick.
     _force_sleep = False
 
-    def __init__(self, time_function=time.perf_counter):
+    def __init__(self, time_function=_time.perf_counter):
         """Initialise a Clock, with optional custom time function.
 
         :Parameters:
@@ -192,8 +176,8 @@ class Clock:
         self.next_ts = self.time()
         self.last_ts = None
 
-        # Used by self.get_fps to show update frequency
-        self.times = deque()
+        # Used by self.get_frequency to show update frequency
+        self.times = _deque()
         self.cumulative_time = 0
         self.window_size = 60
 
@@ -203,7 +187,7 @@ class Clock:
 
     @staticmethod
     def sleep(microseconds):
-        time.sleep(microseconds * 1e-6)
+        _time.sleep(microseconds * 1e-6)
 
     def update_time(self):
         """Get the elapsed time since the last call to `update_time`.
@@ -275,9 +259,9 @@ class Clock:
             # case it needs to be rescheduled.  it is more efficient
             # to push and pop the heap at once rather than two operations
             if item is None:
-                item = heappop(interval_items)
+                item = _heappop(interval_items)
             else:
-                item = heappushpop(interval_items, item)
+                item = _heappushpop(interval_items, item)
 
             # a scheduled function may try and unschedule itself
             # so we need to keep a reference to the current
@@ -289,7 +273,10 @@ class Clock:
                 break
 
             # execute the callback
-            item.func(now - item.last_ts, *item.args, **item.kwargs)
+            try:
+                item.func(now - item.last_ts, *item.args, **item.kwargs)
+            except ReferenceError:
+                pass    # weakly-referenced object no longer exists.
 
             if item.interval:
 
@@ -319,7 +306,7 @@ class Clock:
                 self._current_interval_item = item = None
 
         if item is not None:
-            heappush(interval_items, item)
+            _heappush(interval_items, item)
 
         return True
 
@@ -382,7 +369,7 @@ class Clock:
 
         return None
 
-    def get_fps(self):
+    def get_frequency(self):
         """Get the average clock update frequency of recent history.
 
         The result is the average of a sliding window of the last "n" updates,
@@ -430,7 +417,7 @@ class Clock:
         # and will not always catch smallest value, so sort here.
         # do not remove the sort key...it is faster than relaying comparisons
         # NOTE: do not rewrite as popping from heap, as that is super slow!
-        self._schedule_interval_items.sort(key=attrgetter('next_ts'))
+        self._schedule_interval_items.sort(key=_attrgetter('next_ts'))
 
         # Binary division over interval:
         #
@@ -498,7 +485,7 @@ class Clock:
         last_ts = self._get_nearest_ts()
         next_ts = last_ts + delay
         item = _ScheduledIntervalItem(func, 0, last_ts, next_ts, args, kwargs)
-        heappush(self._schedule_interval_items, item)
+        _heappush(self._schedule_interval_items, item)
 
     def schedule_interval(self, func, interval, *args, **kwargs):
         """Schedule a function to be called every `interval` seconds.
@@ -518,7 +505,7 @@ class Clock:
         last_ts = self._get_nearest_ts()
         next_ts = last_ts + interval
         item = _ScheduledIntervalItem(func, interval, last_ts, next_ts, args, kwargs)
-        heappush(self._schedule_interval_items, item)
+        _heappush(self._schedule_interval_items, item)
 
     def schedule_interval_soft(self, func, interval, *args, **kwargs):
         """Schedule a function to be called every ``interval`` seconds.
@@ -552,7 +539,7 @@ class Clock:
         next_ts = self._get_soft_next_ts(self._get_nearest_ts(), interval)
         last_ts = next_ts - interval
         item = _ScheduledIntervalItem(func, interval, last_ts, next_ts, args, kwargs)
-        heappush(self._schedule_interval_items, item)
+        _heappush(self._schedule_interval_items, item)
 
     def unschedule(self, func):
         """Remove a function from the schedule.
@@ -568,9 +555,7 @@ class Clock:
         # clever remove item without disturbing the heap:
         # 1. set function to an empty lambda -- original function is not called
         # 2. set interval to 0               -- item will be removed from heap eventually
-        valid_items = set(item
-                          for item in self._schedule_interval_items
-                          if item.func == func)
+        valid_items = set(item for item in self._schedule_interval_items if item.func == func)
 
         if self._current_interval_item:
             if self._current_interval_item.func == func:
@@ -653,7 +638,7 @@ def get_sleep_time(sleep_idle):
     return _default.get_sleep_time(sleep_idle)
 
 
-def get_fps():
+def get_frequency():
     """Get the average clock update frequency.
 
     The result is the sliding average of the last "n" updates,
@@ -665,7 +650,7 @@ def get_fps():
     :rtype: float
     :return: The measured updates per second.
     """
-    return _default.get_fps()
+    return _default.get_frequency()
 
 
 def schedule(func, *args, **kwargs):
@@ -722,7 +707,7 @@ def schedule_interval_soft(func, interval, *args, **kwargs):
 def schedule_once(func, delay, *args, **kwargs):
     """Schedule ``func`` to be called once after ``delay`` seconds.
 
-    This function uses the fefault clock. ``delay`` can be a float. The
+    This function uses the default clock. ``delay`` can be a float. The
     arguments passed to ``func`` are ``dt`` (time since last function call),
     followed by any ``*args`` and ``**kwargs`` given here.
 
