@@ -1,7 +1,7 @@
 # ----------------------------------------------------------------------------
 # pyglet
 # Copyright (c) 2006-2008 Alex Holkner
-# Copyright (c) 2008-2020 pyglet contributors
+# Copyright (c) 2008-2022 pyglet contributors
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -63,7 +63,7 @@ properties determine the sprite's :py:attr:`~pyglet.sprite.Sprite.rotation`,
 :py:attr:`~pyglet.sprite.Sprite.scale` and
 :py:attr:`~pyglet.sprite.Sprite.opacity`.
 
-By default sprite coordinates are restricted to integer values to avoid
+By default, sprite coordinates are restricted to integer values to avoid
 sub-pixel artifacts.  If you require to use floats, for example for smoother
 animations, you can set the ``subpixel`` parameter to ``True`` when creating
 the sprite (:since: pyglet 1.2).
@@ -101,7 +101,8 @@ sprites within batches.
 """
 
 import sys
-import math
+
+import pyglet
 
 from pyglet.gl import *
 from pyglet import clock
@@ -112,6 +113,95 @@ from pyglet import image
 _is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run
 
 
+vertex_source = """#version 150 core
+    in vec3 translate;
+    in vec4 colors;
+    in vec3 tex_coords;
+    in vec2 scale;
+    in vec3 position;
+    in float rotation;
+
+    out vec4 vertex_colors;
+    out vec3 texture_coords;
+
+    uniform WindowBlock
+    {
+        mat4 projection;
+        mat4 view;
+    } window;
+
+    mat4 m_scale = mat4(1.0);
+    mat4 m_rotation = mat4(1.0);
+    mat4 m_translate = mat4(1.0);
+
+    void main()
+    {
+        m_scale[0][0] = scale.x;
+        m_scale[1][1] = scale.y;
+        m_translate[3][0] = translate.x;
+        m_translate[3][1] = translate.y;
+        m_translate[3][2] = translate.z;
+        m_rotation[0][0] =  cos(-radians(rotation)); 
+        m_rotation[0][1] =  sin(-radians(rotation));
+        m_rotation[1][0] = -sin(-radians(rotation));
+        m_rotation[1][1] =  cos(-radians(rotation));
+
+        gl_Position = window.projection * window.view * m_translate * m_rotation * m_scale * vec4(position, 1.0);
+
+        vertex_colors = colors;
+        texture_coords = tex_coords;
+    }
+"""
+
+fragment_source = """#version 150 core
+    in vec4 vertex_colors;
+    in vec3 texture_coords;
+    out vec4 final_colors;
+
+    uniform sampler2D sprite_texture;
+
+    void main()
+    {
+        final_colors = texture(sprite_texture, texture_coords.xy) * vertex_colors;
+    }
+"""
+
+fragment_array_source = """#version 150 core
+    in vec4 vertex_colors;
+    in vec3 texture_coords;
+    out vec4 final_colors;
+
+    uniform sampler2DArray sprite_texture;
+
+    void main()
+    {
+        final_colors = texture(sprite_texture, texture_coords) * vertex_colors;
+    }
+"""
+
+
+def get_default_shader():
+    try:
+        return pyglet.gl.current_context.pyglet_sprite_default_shader
+    except AttributeError:
+        _default_vert_shader = graphics.shader.Shader(vertex_source, 'vertex')
+        _default_frag_shader = graphics.shader.Shader(fragment_source, 'fragment')
+        default_shader_program = graphics.shader.ShaderProgram(_default_vert_shader, _default_frag_shader)
+        pyglet.gl.current_context.pyglet_sprite_default_shader = default_shader_program
+        return pyglet.gl.current_context.pyglet_sprite_default_shader
+
+
+def get_default_array_shader():
+    try:
+        return pyglet.gl.current_context.pyglet_sprite_default_array_shader
+    except AttributeError:
+        _default_vert_shader = graphics.shader.Shader(vertex_source, 'vertex')
+        _default_array_frag_shader = graphics.shader.Shader(fragment_array_source, 'fragment')
+        default_shader_program = graphics.shader.ShaderProgram(_default_vert_shader, _default_array_frag_shader)
+        pyglet.gl.current_context.pyglet_sprite_default_array_shader = default_shader_program
+        return pyglet.gl.current_context.pyglet_sprite_default_array_shader
+
+
 class SpriteGroup(graphics.Group):
     """Shared sprite rendering group.
 
@@ -119,11 +209,11 @@ class SpriteGroup(graphics.Group):
     same parent group, texture and blend parameters.
     """
 
-    def __init__(self, texture, blend_src, blend_dest, parent=None):
+    def __init__(self, texture, blend_src, blend_dest, program, order=0, parent=None):
         """Create a sprite group.
 
-        The group is created internally within :py:class:`~pyglet.sprite.Sprite`; applications usually
-        do not need to explicitly create it.
+        The group is created internally when a :py:class:`~pyglet.sprite.Sprite`
+        is created; applications usually do not need to explicitly create it.
 
         :Parameters:
             `texture` : `~pyglet.image.Texture`
@@ -134,39 +224,46 @@ class SpriteGroup(graphics.Group):
             `blend_dest` : int
                 OpenGL blend destination mode; for example,
                 ``GL_ONE_MINUS_SRC_ALPHA``.
+            `program` : `~pyglet.graphics.shader.ShaderProgram`
+                A custom ShaderProgram.
+            `order` : int
+                Change the order to render above or below other Groups.
             `parent` : `~pyglet.graphics.Group`
                 Optional parent group.
         """
-        super(SpriteGroup, self).__init__(parent)
+        super().__init__(order, parent)
         self.texture = texture
         self.blend_src = blend_src
         self.blend_dest = blend_dest
+        self.program = program
 
     def set_state(self):
-        glEnable(self.texture.target)
+        self.program.use()
+
+        glActiveTexture(GL_TEXTURE0)
         glBindTexture(self.texture.target, self.texture.id)
 
-        glPushAttrib(GL_COLOR_BUFFER_BIT)
         glEnable(GL_BLEND)
         glBlendFunc(self.blend_src, self.blend_dest)
 
     def unset_state(self):
-        glPopAttrib()
-        glDisable(self.texture.target)
+        glDisable(GL_BLEND)
+        self.program.stop()
 
     def __repr__(self):
-        return '%s(%r)' % (self.__class__.__name__, self.texture)
+        return "{0}({1})".format(self.__class__.__name__, self.texture)
 
     def __eq__(self, other):
         return (other.__class__ is self.__class__ and
-                self.parent is other.parent and
+                self.program is other.program and
+                self.parent == other.parent and
                 self.texture.target == other.texture.target and
                 self.texture.id == other.texture.id and
                 self.blend_src == other.blend_src and
                 self.blend_dest == other.blend_dest)
 
     def __hash__(self):
-        return hash((id(self.parent),
+        return hash((id(self.parent), id(self.program),
                      self.texture.id, self.texture.target,
                      self.blend_src, self.blend_dest))
 
@@ -179,6 +276,8 @@ class Sprite(event.EventDispatcher):
 
     _batch = None
     _animation = None
+    _frame_index = 0
+    _paused = False
     _rotation = 0
     _opacity = 255
     _rgb = (255, 255, 255)
@@ -187,14 +286,14 @@ class Sprite(event.EventDispatcher):
     _scale_y = 1.0
     _visible = True
     _vertex_list = None
+    group_class = SpriteGroup
 
     def __init__(self,
-                 img, x=0, y=0,
+                 img, x=0, y=0, z=0,
                  blend_src=GL_SRC_ALPHA,
                  blend_dest=GL_ONE_MINUS_SRC_ALPHA,
                  batch=None,
                  group=None,
-                 usage='dynamic',
                  subpixel=False):
         """Create a sprite.
 
@@ -205,6 +304,8 @@ class Sprite(event.EventDispatcher):
                 X coordinate of the sprite.
             `y` : int
                 Y coordinate of the sprite.
+            `z` : int
+                Z coordinate of the sprite.
             `blend_src` : int
                 OpenGL blend source mode.  The default is suitable for
                 compositing sprites drawn from back-to-front.
@@ -215,23 +316,17 @@ class Sprite(event.EventDispatcher):
                 Optional batch to add the sprite to.
             `group` : `~pyglet.graphics.Group`
                 Optional parent group of the sprite.
-            `usage` : str
-                Vertex buffer object usage hint, one of ``"none"``,
-                ``"stream"``, ``"dynamic"`` (default) or ``"static"``.  Applies
-                only to vertex data.
             `subpixel` : bool
                 Allow floating-point coordinates for the sprite. By default,
                 coordinates are restricted to integer values.
         """
-        if batch is not None:
-            self._batch = batch
-
         self._x = x
         self._y = y
+        self._z = z
+        self._img = img
 
         if isinstance(img, image.Animation):
             self._animation = img
-            self._frame_index = 0
             self._texture = img.frames[0].image.get_texture()
             self._next_dt = img.frames[0].duration
             if self._next_dt:
@@ -239,11 +334,20 @@ class Sprite(event.EventDispatcher):
         else:
             self._texture = img.get_texture()
 
-        self._group = SpriteGroup(self._texture, blend_src, blend_dest, group)
-        self._usage = usage
+        self._batch = batch or graphics.get_default_batch()
+        self._group = self.group_class(self._texture, blend_src, blend_dest, self.program, 0, group)
         self._subpixel = subpixel
         self._create_vertex_list()
 
+    @property
+    def program(self):
+        if isinstance(self._img, image.TextureArrayRegion):
+            program = get_default_array_shader()
+        else:
+            program = get_default_shader()
+
+        return program
+
     def __del__(self):
         try:
             if self._vertex_list is not None:
@@ -304,7 +408,7 @@ class Sprite(event.EventDispatcher):
             return
 
         if batch is not None and self._batch is not None:
-            self._batch.migrate(self._vertex_list, GL_QUADS, self._group, batch)
+            self._batch.migrate(self._vertex_list, GL_TRIANGLES, self._group, batch)
             self._batch = batch
         else:
             self._vertex_list.delete()
@@ -326,13 +430,13 @@ class Sprite(event.EventDispatcher):
     def group(self, group):
         if self._group.parent == group:
             return
-        self._group = SpriteGroup(self._texture,
-                                  self._group.blend_src,
-                                  self._group.blend_dest,
-                                  group)
-        if self._batch is not None:
-            self._batch.migrate(self._vertex_list, GL_QUADS, self._group,
-                                self._batch)
+        self._group = self.group_class(self._texture,
+                                       self._group.blend_src,
+                                       self._group.blend_dest,
+                                       self._group.program,
+                                       0,
+                                       group)
+        self._batch.migrate(self._vertex_list, GL_TRIANGLES, self._group, self._batch)
 
     @property
     def image(self):
@@ -364,99 +468,63 @@ class Sprite(event.EventDispatcher):
 
     def _set_texture(self, texture):
         if texture.id is not self._texture.id:
-            self._group = SpriteGroup(texture,
-                                      self._group.blend_src,
-                                      self._group.blend_dest,
-                                      self._group.parent)
-            if self._batch is None:
-                self._vertex_list.tex_coords[:] = texture.tex_coords
-            else:
-                self._vertex_list.delete()
-                self._texture = texture
-                self._create_vertex_list()
+            self._group = self._group.__class__(texture,
+                                                self._group.blend_src,
+                                                self._group.blend_dest,
+                                                self._group.program,
+                                                0,
+                                                self._group.parent)
+            self._vertex_list.delete()
+            self._texture = texture
+            self._create_vertex_list()
         else:
             self._vertex_list.tex_coords[:] = texture.tex_coords
         self._texture = texture
 
     def _create_vertex_list(self):
-        if self._subpixel:
-            vertex_format = 'v2f/%s' % self._usage
-        else:
-            vertex_format = 'v2i/%s' % self._usage
-        if self._batch is None:
-            self._vertex_list = graphics.vertex_list(
-                4, vertex_format, 'c4B', ('t3f', self._texture.tex_coords))
-        else:
-            self._vertex_list = self._batch.add(
-                4, GL_QUADS, self._group, vertex_format, 'c4B', ('t3f', self._texture.tex_coords))
+        self._vertex_list = self.program.vertex_list_indexed(
+            4, GL_TRIANGLES, [0, 1, 2, 0, 2, 3], self._batch, self._group,
+            colors=('Bn', (*self._rgb, int(self._opacity)) * 4),
+            translate=('f', (self._x, self._y, self._z) * 4),
+            scale=('f', (self._scale*self._scale_x, self._scale*self._scale_y) * 4),
+            rotation=('f', (self._rotation,) * 4),
+            tex_coords=('f', self._texture.tex_coords))
         self._update_position()
-        self._update_color()
 
     def _update_position(self):
-        img = self._texture
-        scale_x = self._scale * self._scale_x
-        scale_y = self._scale * self._scale_y
         if not self._visible:
-            vertices = (0, 0, 0, 0, 0, 0, 0, 0)
-        elif self._rotation:
-            x1 = -img.anchor_x * scale_x
-            y1 = -img.anchor_y * scale_y
-            x2 = x1 + img.width * scale_x
-            y2 = y1 + img.height * scale_y
-            x = self._x
-            y = self._y
-
-            r = -math.radians(self._rotation)
-            cr = math.cos(r)
-            sr = math.sin(r)
-            ax = x1 * cr - y1 * sr + x
-            ay = x1 * sr + y1 * cr + y
-            bx = x2 * cr - y1 * sr + x
-            by = x2 * sr + y1 * cr + y
-            cx = x2 * cr - y2 * sr + x
-            cy = x2 * sr + y2 * cr + y
-            dx = x1 * cr - y2 * sr + x
-            dy = x1 * sr + y2 * cr + y
-            vertices = (ax, ay, bx, by, cx, cy, dx, dy)
-        elif scale_x != 1.0 or scale_y != 1.0:
-            x1 = self._x - img.anchor_x * scale_x
-            y1 = self._y - img.anchor_y * scale_y
-            x2 = x1 + img.width * scale_x
-            y2 = y1 + img.height * scale_y
-            vertices = (x1, y1, x2, y1, x2, y2, x1, y2)
+            self._vertex_list.position[:] = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
         else:
-            x1 = self._x - img.anchor_x
-            y1 = self._y - img.anchor_y
+            img = self._texture
+            x1 = -img.anchor_x
+            y1 = -img.anchor_y
             x2 = x1 + img.width
             y2 = y1 + img.height
-            vertices = (x1, y1, x2, y1, x2, y2, x1, y2)
-        if not self._subpixel:
-            vertices = (int(vertices[0]), int(vertices[1]),
-                        int(vertices[2]), int(vertices[3]),
-                        int(vertices[4]), int(vertices[5]),
-                        int(vertices[6]), int(vertices[7]))
-        self._vertex_list.vertices[:] = vertices
+            vertices = (x1, y1, 0, x2, y1, 0, x2, y2, 0, x1, y2, 0)
 
-    def _update_color(self):
-        r, g, b = self._rgb
-        self._vertex_list.colors[:] = [r, g, b, int(self._opacity)] * 4
+            if not self._subpixel:
+                self._vertex_list.position[:] = tuple(map(int, vertices))
+            else:
+                self._vertex_list.position[:] = vertices
 
     @property
     def position(self):
-        """The (x, y) coordinates of the sprite, as a tuple.
+        """The (x, y, z) coordinates of the sprite, as a tuple.
 
         :Parameters:
             `x` : int
                 X coordinate of the sprite.
             `y` : int
                 Y coordinate of the sprite.
+            `z` : int
+                Z coordinate of the sprite.
         """
-        return self._x, self._y
+        return self._x, self._y, self._z
 
     @position.setter
-    def position(self, pos):
-        self._x, self._y = pos
-        self._update_position()
+    def position(self, position):
+        self._x, self._y, self._z = position
+        self._vertex_list.translate[:] = position * 4
 
     @property
     def x(self):
@@ -469,7 +537,7 @@ class Sprite(event.EventDispatcher):
     @x.setter
     def x(self, x):
         self._x = x
-        self._update_position()
+        self._vertex_list.translate[:] = (x, self._y, self._z) * 4
 
     @property
     def y(self):
@@ -482,7 +550,20 @@ class Sprite(event.EventDispatcher):
     @y.setter
     def y(self, y):
         self._y = y
-        self._update_position()
+        self._vertex_list.translate[:] = (self._x, y, self._z) * 4
+
+    @property
+    def z(self):
+        """Z coordinate of the sprite.
+
+        :type: int
+        """
+        return self._z
+
+    @z.setter
+    def z(self, z):
+        self._z = z
+        self._vertex_list.translate[:] = (self._x, self._y, z) * 4
 
     @property
     def rotation(self):
@@ -498,7 +579,7 @@ class Sprite(event.EventDispatcher):
     @rotation.setter
     def rotation(self, rotation):
         self._rotation = rotation
-        self._update_position()
+        self._vertex_list.rotation[:] = (self._rotation,) * 4
 
     @property
     def scale(self):
@@ -514,12 +595,12 @@ class Sprite(event.EventDispatcher):
     @scale.setter
     def scale(self, scale):
         self._scale = scale
-        self._update_position()
+        self._vertex_list.scale[:] = (scale * self._scale_x, scale * self._scale_y) * 4
 
     @property
     def scale_x(self):
         """Horizontal scaling factor.
-        
+
          A scaling factor of 1 (the default) has no effect.  A scale of 2 will
          draw the sprite at twice the native width of its image.
 
@@ -530,12 +611,12 @@ class Sprite(event.EventDispatcher):
     @scale_x.setter
     def scale_x(self, scale_x):
         self._scale_x = scale_x
-        self._update_position()
+        self._vertex_list.scale[:] = (self._scale * scale_x, self._scale * self._scale_y) * 4
 
     @property
     def scale_y(self):
         """Vertical scaling factor.
-        
+
          A scaling factor of 1 (the default) has no effect.  A scale of 2 will
          draw the sprite at twice the native height of its image.
 
@@ -546,21 +627,21 @@ class Sprite(event.EventDispatcher):
     @scale_y.setter
     def scale_y(self, scale_y):
         self._scale_y = scale_y
-        self._update_position()
+        self._vertex_list.scale[:] = (self._scale * self._scale_x, self._scale * scale_y) * 4
 
-    def update(self, x=None, y=None, rotation=None, scale=None, scale_x=None, scale_y=None):
+    def update(self, x=None, y=None, z=None, rotation=None, scale=None, scale_x=None, scale_y=None):
         """Simultaneously change the position, rotation or scale.
 
-        This method is provided for performance. In cases where
-        multiple Sprite attributes need to be updated at the same
-        time, it is more efficent to update them together using
-        the update method, rather than modifying them one by one.
+        This method is provided for convenience. There is not much
+        performance benefit to updating multiple Sprite attributes at once.
 
         :Parameters:
             `x` : int
                 X coordinate of the sprite.
             `y` : int
                 Y coordinate of the sprite.
+            `z` : int
+                Z coordinate of the sprite.
             `rotation` : float
                 Clockwise rotation of the sprite, in degrees.
             `scale` : float
@@ -570,19 +651,42 @@ class Sprite(event.EventDispatcher):
             `scale_y` : float
                 Vertical scaling factor.
         """
+
+        translations_outdated = False
+
+        # only bother updating if the translation actually changed
         if x is not None:
             self._x = x
+            translations_outdated = True
         if y is not None:
             self._y = y
-        if rotation is not None:
+            translations_outdated = True
+        if z is not None:
+            self._z = z
+            translations_outdated = True
+
+        if translations_outdated:
+            self._vertex_list.translate[:] = (self._x, self._y, self._z) * 4
+
+        if rotation is not None and rotation != self._rotation:
             self._rotation = rotation
+            self._vertex_list.rotation[:] = (rotation,) * 4
+
+        scales_outdated = False
+
+        # only bother updating if the scale actually changed
         if scale is not None:
             self._scale = scale
+            scales_outdated = True
         if scale_x is not None:
             self._scale_x = scale_x
+            scales_outdated = True
         if scale_y is not None:
             self._scale_y = scale_y
-        self._update_position()
+            scales_outdated = True
+
+        if scales_outdated:
+            self._vertex_list.scale[:] = (self._scale * self._scale_x, self._scale * self._scale_y) * 4
 
     @property
     def width(self):
@@ -629,7 +733,7 @@ class Sprite(event.EventDispatcher):
     @opacity.setter
     def opacity(self, opacity):
         self._opacity = opacity
-        self._update_color()
+        self._vertex_list.colors[:] = (*self._rgb, int(self._opacity)) * 4
 
     @property
     def color(self):
@@ -648,7 +752,7 @@ class Sprite(event.EventDispatcher):
     @color.setter
     def color(self, rgb):
         self._rgb = list(map(int, rgb))
-        self._update_color()
+        self._vertex_list.colors[:] = (*self._rgb, int(self._opacity)) * 4
 
     @property
     def visible(self):
@@ -663,6 +767,51 @@ class Sprite(event.EventDispatcher):
         self._visible = visible
         self._update_position()
 
+    @property
+    def paused(self):
+        """Pause/resume the Sprite's Animation
+
+        If `Sprite.image` is an Animation, you can pause or resume
+        the animation by setting this property to True or False.
+        If not an Animation, this has no effect.
+
+        :type: bool
+        """
+        return self._paused
+
+    @paused.setter
+    def paused(self, pause):
+        if not hasattr(self, '_animation') or pause == self._paused:
+            return
+        if pause is True:
+            clock.unschedule(self._animate)
+        else:
+            frame = self._animation.frames[self._frame_index]
+            self._next_dt = frame.duration
+            if self._next_dt:
+                clock.schedule_once(self._animate, self._next_dt)
+        self._paused = pause
+
+    @property
+    def frame_index(self):
+        """The current Animation frame.
+
+        If the `Sprite.image` is an `Animation`,
+        you can query or set the current frame.
+        If not an Animation, this will always
+        be 0.
+
+        :type: int
+        """
+        return self._frame_index
+
+    @frame_index.setter
+    def frame_index(self, index):
+        # Bound to available number of frames
+        if self._animation is None:
+            return
+        self._frame_index = max(0, min(index, len(self._animation.frames)-1))
+
     def draw(self):
         """Draw the sprite at its current position.
 
@@ -670,7 +819,7 @@ class Sprite(event.EventDispatcher):
         efficiently.
         """
         self._group.set_state_recursive()
-        self._vertex_list.draw(GL_QUADS)
+        self._vertex_list.draw(GL_TRIANGLES)
         self._group.unset_state_recursive()
 
     if _is_pyglet_doc_run:
@@ -686,3 +835,47 @@ class Sprite(event.EventDispatcher):
 
 
 Sprite.register_event_type('on_animation_end')
+
+
+class AdvancedSprite(pyglet.sprite.Sprite):
+    """Is a sprite that lets you change the shader program during initialization and after
+    For advanced users who understand shaders."""
+    def __init__(self,
+                 img, x=0, y=0, z=0,
+                 blend_src=GL_SRC_ALPHA,
+                 blend_dest=GL_ONE_MINUS_SRC_ALPHA,
+                 batch=None,
+                 group=None,
+                 subpixel=False,
+                 program=None):
+
+        self._program = program
+
+        if not program:
+            if isinstance(img, image.TextureArrayRegion):
+                self._program = get_default_array_shader()
+            else:
+                self._program = get_default_shader()
+
+        super().__init__(img, x, y, z, blend_src, blend_dest, batch, group, subpixel)
+
+    @property
+    def program(self):
+        return self._program
+
+    @program.setter
+    def program(self, program):
+        if self._program == program:
+            return
+        self._group = self.group_class(self._texture,
+                                       self._group.blend_src,
+                                       self._group.blend_dest,
+                                       program,
+                                       0,
+                                       self._group)
+        self._batch.migrate(self._vertex_list, GL_TRIANGLES, self._group, self._batch)
+        self._program = program
+
+
+
+
